import os
import pickle
from typing import Tuple

import pandas as pd
from google.auth.transport.requests import Request
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build
from nure.sync.cache import LocalFileCache


class GoogleSheet(LocalFileCache):
    def __init__(self, key_file, credential_file, root_path='data/googlesheet', ttl=None) -> None:
        super().__init__(root_path, ttl=ttl)
        self.key_file = key_file
        self.credential_file = credential_file

    def _parse_key_(key: dict) -> Tuple[str, str]:
        if not isinstance(key, dict):
            raise TypeError('key must be a dict')

        spreadsheet_id = key['spreadsheet_id']
        range_ = key['range']
        return spreadsheet_id, range_

    def key_to_local_relative_path(self, key, *args, **kargs) -> str:
        spreadsheet_id, range_ = GoogleSheet._parse_key_(key)
        components = range_.split('!')
        if len(components) == 1:
            # sheet_name or cell range
            if len(components[0].split(':')) == 2:
                sheet_name = None
            else:
                sheet_name = components[0]
        else:
            sheet_name = '!'.join(components[:-1])

        if sheet_name is None:
            return f'{spreadsheet_id}.csv'

        if sheet_name.startswith("'") and sheet_name.endswith("'"):
            sheet_name = sheet_name[1:-1]

        relative_path = os.path.join(spreadsheet_id, f'{sheet_name}.csv')
        return relative_path

    def retrieve(self, key, local_file_path, header=False, *args, **kargs) -> None:
        """Shows basic usage of the Sheets API.
        Prints values from a sample spreadsheet.
        """
        spreadsheet_id, range_ = GoogleSheet._parse_key_(key)
        scopes = ['https://www.googleapis.com/auth/spreadsheets.readonly']
        creds = None
        # The file token.pickle stores the user's access and refresh tokens, and is
        # created automatically when the authorization flow completes for the first time.
        if os.path.exists(self.key_file):
            with open(self.key_file, 'rb') as token:
                creds = pickle.load(token)
        # If there are no (valid) credentials available, let the user log in.
        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                creds.refresh(Request())
            else:
                flow = InstalledAppFlow.from_client_secrets_file(self.credential_file, scopes)
                creds = flow.run_local_server(port=0)
            # Save the credentials for the next run
            with open(self.key_file, 'wb') as token:
                pickle.dump(creds, token)

        service = build('sheets', 'v4', credentials=creds)

        # Call the Sheets API
        sheet = service.spreadsheets()
        result = sheet.values().get(
            spreadsheetId=spreadsheet_id,
            range=range_).execute()

        values = result['values']

        columns = header
        if header is True:
            columns = values[0]
            values = values[1:]

        pd_data = pd.DataFrame(data=result['values'], columns=columns)
        pd_data.to_csv(local_file_path, index=False)



import os


def splitall(path):
    allparts = []
    while True:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break
        elif parts[1] == path:  # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            allparts.insert(0, parts[1])
    return allparts



import sqlalchemy
import nure.sync.sql


class BigQuery(nure.sync.sql.Sql):
    def __init__(self, keyfile, suffix_func, arraysize=10_000, root_path='data/bigquery', ttl=None) -> None:
        super(BigQuery, self).__init__(suffix_func, root_path, ttl)

        self._engine = sqlalchemy.create_engine('bigquery://', credentials_path=keyfile, arraysize=arraysize)

    @property
    def engine(self) -> sqlalchemy.engine.Engine:
        return self._engine



import json
from datetime import datetime, date
from collections.abc import Iterable


class DatetimeJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime) or isinstance(obj, date):
            return obj.isoformat()
        return json.JSONEncoder.default(self, obj)


def parse(obj):
    if isinstance(obj, str):
        try:
            obj = datetime.fromisoformat(obj)
        except ValueError:
            pass
    elif isinstance(obj, dict):
        obj = {key: parse(value) for key, value in obj.items()}
    elif isinstance(obj, list):
        obj = [parse(item) for item in obj]
    elif isinstance(obj, tuple):
        obj = tuple([parse(item) for item in obj])
    elif isinstance(obj, Iterable):
        obj = map(parse, obj)

    return obj


def to_string(obj):
    if isinstance(obj, datetime) or isinstance(obj, date):
        return obj.isoformat()
    elif isinstance(obj, dict):
        obj = {key: to_string(value) for key, value in obj.items()}
    elif isinstance(obj, list):
        obj = [to_string(item) for item in obj]
    elif isinstance(obj, tuple):
        obj = tuple([to_string(item) for item in obj])
    elif isinstance(obj, Iterable):
        obj = map(to_string, obj)

    return obj



import os
import re
from typing import Union

import boto3
import nure.path
import nure.sync.cache
import botocore.exceptions


class S3Uri:
    def __init__(self, bucket: str, key: str) -> None:
        self.bucket = bucket
        self.key = key

    PARSE_RE = re.compile(R"^s3://(?P<bucket>[0-9a-z.-]+)\/(?P<key>([0-9A-Za-z!\-_.*'\()]+/?)+)(?<!/)$")

    @staticmethod
    def parse(uri: str):
        # Bucket name
        # https://docs.aws.amazon.com/awscloudtrail/latest/userguide/cloudtrail-s3-bucket-naming-requirements.html
        # Object name
        # https://docs.aws.amazon.com/AmazonS3/latest/userguide/object-keys.html
        # this implementation just covers sub-set of S3 URI
        match = S3Uri.PARSE_RE.match(uri)
        if match is None:
            raise ValueError('input string is not S3 URI')

        group_dict = match.groupdict()
        return S3Uri(group_dict['bucket'], group_dict['key'])

    @property
    def uri(self):
        return f's3://{self.bucket}/{self.key}'

    def __str__(self) -> str:
        return self.uri

    def __repr__(self) -> str:
        return f'S3Uri({self.bucket}, {self.key})'


class S3(nure.sync.cache.LocalFileCache):
    def __init__(self, root_path='data/s3', ttl=None) -> None:
        super(S3, self).__init__(root_path, ttl)

    @staticmethod
    def _to_s3uri_(s3_uri: Union[str, S3Uri]):
        # unify s3_uri type
        if isinstance(s3_uri, str):
            s3_uri = S3Uri.parse(s3_uri)

        return s3_uri

    def key_to_local_relative_path(self, s3_uri: Union[str, S3Uri], *args, **kargs) -> str:
        s3_uri = S3._to_s3uri_(s3_uri)
        return os.path.join(s3_uri.bucket, s3_uri.key)

    def retrieve(self, s3_uri: Union[str, S3Uri], local_file_path):
        s3_uri = S3._to_s3uri_(s3_uri)

        s3 = boto3.resource('s3')
        s3.meta.client.download_file(s3_uri.bucket, s3_uri.key, local_file_path)

    def _infer_s3uri_(self, local_file_path):
        relative_path = os.path.relpath(local_file_path, self.root_path)
        components = nure.path.splitall(relative_path)
        bucket = components[0]
        key = '/'.join(components[1:])

        return S3Uri(bucket, key)

    def upload(self, local_file_path, s3_uri=None):
        if s3_uri is None:
            s3_uri = self._infer_s3uri_(local_file_path)
        else:
            s3_uri = S3._to_s3uri_(s3_uri)

        s3 = boto3.resource('s3')
        s3.meta.client.upload_file(local_file_path, s3_uri.bucket, s3_uri.key)

        return s3_uri

    def metadata(self, s3_uri: Union[str, S3Uri]):
        if isinstance(s3_uri, str):
            s3_uri = S3Uri.parse(s3_uri)

        s3 = boto3.resource('s3')
        try:
            meta = s3.meta.client.head_object(Bucket=s3_uri.bucket, Key=s3_uri.key)
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == '404':
                return None

            raise e

        return meta

    def exist(self, s3_uri: Union[str, S3Uri]):
        return self.metadata(s3_uri) is not None



import glob
import os
import shutil
import tempfile
import zipfile


class Archive:
    def __init__(self, filename: str, root_dir: str = None,
                 compression=zipfile.ZIP_DEFLATED) -> None:
        self.filename = filename
        self.root_dir = root_dir
        self.compression = compression
        self.ensure_dirs()

    def ensure_dirs(self):
        self.use_tempfile = self.root_dir is None
        if self.use_tempfile:
            self.root_dir = tempfile.mkdtemp()

        self.org_dir = os.path.join(self.root_dir, 'original')
        self.mod_dir = os.path.join(self.root_dir, 'modified')
        os.makedirs(self.org_dir, exist_ok=True)
        os.makedirs(self.mod_dir, exist_ok=True)

    def cleanup(self, temp_only=False):
        if self.use_tempfile:
            shutil.rmtree(self.root_dir)
            self.root_dir = None
        elif not temp_only:
            shutil.rmtree(self.org_dir)
            shutil.rmtree(self.mod_dir)

    def __del__(self):
        self.cleanup(temp_only=True)

    def unpack(self, force=False):
        if force or (not any(os.scandir(self.org_dir))):
            shutil.unpack_archive(self.filename, self.org_dir)

    def get(self, name: str, mode='o'):
        if mode == 'o':
            return os.path.join(self.org_dir, name)
        elif mode == 'm':
            return os.path.join(self.mod_dir, name)
        elif mode == 'c':
            mod_path = os.path.join(self.mod_dir, name)
            if os.path.exists(mod_path):
                return mod_path
            else:
                return os.path.join(self.org_dir, name)

        raise ValueError("mode must be 'o' (original), 'm' (modified) or 'c' (current)")

    def pack(self):
        with zipfile.ZipFile(self.filename, mode='w', compression=self.compression) as zf:
            # add non-modified element
            for path in glob.iglob(f'{self.org_dir}/**', recursive=True):
                if not os.path.isfile(path):
                    continue

                arcname = os.path.relpath(path, self.org_dir)
                # if is modified
                if os.path.isfile(os.path.join(self.mod_dir, arcname)):
                    continue

                zf.write(path, arcname)

            # add modifed element
            for path in glob.iglob(f'{self.mod_dir}/**', recursive=True):
                if not os.path.isfile(path):
                    continue

                arcname = os.path.relpath(path, self.mod_dir)
                zf.write(path, arcname)



from typing import Any, Callable, Iterable


class Chain:
    def __init__(self, *l_func: Iterable[Callable]) -> None:
        self.l_func = l_func

    def __call__(self, argument: Any) -> Any:
        for func in self.l_func:
            argument = func(argument)

        return argument


class FromIterable:
    def __init__(self, func: Callable) -> None:
        self.func = func

    def __call__(self, iterable_arguments: Iterable):
        return self.func(*iterable_arguments)



import concurrent.futures
import os
from collections import Iterable, deque
from typing import Callable, List

_DEFAULT_N_WORKERS_ = os.cpu_count() or 4


def parallelize_iterable(iterable: Iterable, func: Callable, *args, preserve_order=True,
                         executor: concurrent.futures.Executor = None,
                         n_workers: int = _DEFAULT_N_WORKERS_, use_process: bool = False,
                         **kwargs):
    # check if exe is provided
    if executor is None:
        if use_process:
            my_executor = concurrent.futures.ProcessPoolExecutor(max_workers=n_workers)
        else:
            my_executor = concurrent.futures.ThreadPoolExecutor(max_workers=n_workers)
    else:
        my_executor = executor

    # submit tasks
    task_list = deque([my_executor.submit(func, item, *args, **kwargs) for item in iterable])

    # collect results
    # note to remove completed task from the list for memory efficiency
    for task in concurrent.futures.as_completed(task_list):
        if preserve_order:
            while len(task_list) > 0 and task_list[0].done():
                yield task_list.popleft().result()
        else:
            task_list.remove(task)
            yield task.result()

            # if is my exe, then shutdown
    if executor is None:
        my_executor.shutdown()


def apply_cache(l_key: List, d_cache: dict, func: Callable):
    l_result = [None] * len(l_key)
    l_miss_index = []
    l_miss_key = []

    for index, key in enumerate(l_key):
        if key in d_cache:
            l_result[index] = d_cache[key]
        else:
            l_miss_index.append(index)
            l_miss_key.append(key)

    l_miss_result = func(l_miss_key)
    # add to result and add to cache
    for index, key, result in zip(l_miss_index, l_miss_key, l_miss_result):
        l_result[index] = result
        d_cache[key] = result

    return l_result


def get_as_iterable(arr: List, indices: List[int]):
    for index in indices:
        yield arr[index]


def get_many(arr: List, indices: List[int]):
    return list(get_as_iterable(arr, indices))


def set_many(arr: List, indices: List[int], values: Iterable):
    for index, value in zip(indices, values):
        arr[index] = value


def unique(arr: List):
    seen = set()
    seen_add = seen.add
    return [x for x in arr if not (x in seen or seen_add(x))]



import collections.abc


def update(default, modified):
    for k, v in modified.items():
        if isinstance(v, collections.abc.Mapping):
            default[k] = update(default.get(k, {}), v)
        else:
            default[k] = v
    return default


def get(obj, key, sep='.', default=None, suppress_error=False):
    l_key = key.split(sep)
    for k in l_key:
        if not isinstance(obj, dict) or k not in obj:
            if suppress_error:
                return default

            raise KeyError(f'{k} is not in {obj}')
        obj = obj[k]

    return obj


def set(obj, key, value, sep='.'):
    l_key = key.split(sep)
    for k in l_key[:-1]:
        assert isinstance(obj, dict), ValueError(f'{obj} is not a dict')
        if k not in obj:
            obj[k] = dict()
        obj = obj[k]

    obj[l_key[-1]] = value


def delete(obj, key, sep='.', suppress_error=False):
    l_key = key.split(sep)
    pre_obj = obj
    for k in l_key:
        if not isinstance(obj, dict) or k not in obj:
            if suppress_error:
                return None

            raise ValueError(f'{k} is not in {obj}')

        pre_obj = obj
        obj = obj[k]

    value = pre_obj[l_key[-1]]
    del pre_obj[l_key[-1]]
    return value
